from .constants import (
    ARCHIVE_TAR_EXTENSIONS,
    ARCHIVE_ZIP_EXTENSIONS,
    TEMPORARY_FOLDER)
from .exceptions import (
    BadArchiveError,
    FileExtensionError,
    InvisibleRoadsMacrosDiskError,
    PathValidationError)
from .operations import (
    TemporaryStorage,
    archive_safely,
    archive_tar_safely,
    archive_zip_safely,
    cache_download,
    make_enumerated_folder,
    make_folder,
    make_random_folder,
    make_unique_folder,
    remove_folder,
    remove_path,
    unarchive_safely)
from .resolutions import (
    check_absolute_path,
    check_path,
    check_relative_path,
    get_absolute_path,
    get_asset_path,
    get_file_hash,
    get_real_path,
    get_relative_path,
    has_extension,
    is_matching_path,
    is_path_in_folder,
    walk_paths)


# flake8: noqa: E401



import tarfile
from invisibleroads_macros_disk import (
    BadArchiveError,
    FileExtensionError,
    InvisibleRoadsMacrosDiskError,
    TemporaryStorage,
    archive_safely,
    archive_tar_safely,
    archive_zip_safely,
    make_enumerated_folder,
    make_folder,
    make_random_folder,
    remove_folder,
    remove_path,
    unarchive_safely,
    ARCHIVE_TAR_EXTENSIONS,
    ARCHIVE_ZIP_EXTENSIONS)
from invisibleroads_macros_security import ALPHABET
from os.path import basename, exists, join
from pytest import raises
from shutil import make_archive
from tempfile import mkstemp
from zipfile import ZipFile

from conftest import EXAMPLES_FOLDER, FILE_CONTENT, FILE_NAME


def test_archive_safely(tmpdir):
    source_folder = tmpdir.mkdir('x')
    source_folder.join(FILE_NAME).write(FILE_CONTENT)
    source_folder.join(FILE_NAME + '_').write(FILE_CONTENT)
    source_folder = source_folder.strpath

    archive_path = archive_safely(source_folder, excluded_paths=['*.txt'])
    archive_file = ZipFile(archive_path)
    file_paths = archive_file.namelist()
    assert FILE_NAME not in file_paths
    assert FILE_NAME + '_' in file_paths

    archive_path = archive_safely(EXAMPLES_FOLDER, tmpdir.join(
        'examples.tar.gz').strpath)
    archive_file = tarfile.open(archive_path, 'r:gz')
    file_paths = archive_file.getnames()
    assert join('a', FILE_NAME) in file_paths
    assert join('b', FILE_NAME) not in file_paths

    with raises(FileExtensionError):
        archive_safely(source_folder, source_folder + '.x')


def test_archive_zip_safely(tmpdir):
    source_folder = tmpdir.mkdir('x')
    source_folder.join(FILE_NAME).write(FILE_CONTENT)
    source_folder = source_folder.strpath

    archive_path = archive_zip_safely(source_folder)
    archive_file = ZipFile(archive_path)
    assert FILE_NAME in archive_file.namelist()

    with raises(FileExtensionError):
        archive_zip_safely(
            source_folder, source_folder + ARCHIVE_TAR_EXTENSIONS[0])


def test_archive_tar_safely(tmpdir):
    source_folder = tmpdir.mkdir('x')
    source_folder.join(FILE_NAME).write(FILE_CONTENT)
    source_folder = source_folder.strpath

    archive_path = archive_tar_safely(source_folder)
    archive_file = tarfile.open(archive_path)
    assert FILE_NAME in archive_file.getnames()

    with raises(FileExtensionError):
        archive_tar_safely(
            source_folder, source_folder + ARCHIVE_ZIP_EXTENSIONS[0])


def test_unarchive_safely(tmpdir):
    with raises(FileExtensionError):
        unarchive_safely(tmpdir.join('x.x').strpath)
    with raises(BadArchiveError):
        archive_path_object = tmpdir.join('x.zip')
        archive_path_object.write(FILE_CONTENT)
        unarchive_safely(archive_path_object.strpath)
    with raises(BadArchiveError):
        archive_path_object = tmpdir.join('x.tar.gz')
        archive_path_object.write(FILE_CONTENT)
        unarchive_safely(archive_path_object.strpath)

    archive_path = EXAMPLES_FOLDER + '.zip'
    archive_file = ZipFile(archive_path)
    assert 'b/file.txt' in archive_file.namelist()
    archive_folder = unarchive_safely(archive_path, tmpdir.join(
        'examples').strpath)
    assert open(join(archive_folder, 'a', FILE_NAME)).read() == FILE_CONTENT
    assert not exists(join(archive_folder, 'b', FILE_NAME))

    archive_basename = tmpdir.join('examples').strpath
    make_archive(archive_basename, 'gztar', EXAMPLES_FOLDER)
    archive_path = archive_basename + '.tar.gz'
    archive_folder = unarchive_safely(archive_path)
    assert open(join(archive_folder, 'a', FILE_NAME)).read() == FILE_CONTENT
    assert not exists(join(archive_folder, 'b', FILE_NAME))


def test_temporary_storage():
    with TemporaryStorage() as storage:
        assert exists(storage.folder)
    assert not exists(storage.folder)


def test_make_enumerated_folder(tmpdir):
    folder = make_enumerated_folder(tmpdir)
    assert basename(folder) == '1'
    folder = make_enumerated_folder(tmpdir)
    assert basename(folder) == '2'


def test_make_random_folder(tmpdir):
    for i in range(len(ALPHABET) + 1):
        folder = make_random_folder(tmpdir, 1)
        assert len(basename(folder)) >= 1
    for x in ALPHABET:
        make_folder(tmpdir.join(x).strpath)
    with raises(InvisibleRoadsMacrosDiskError):
        folder = make_random_folder(tmpdir, 1, with_fixed_length=True)


def test_remove_folder():
    temporary_folder = make_random_folder()
    remove_folder(temporary_folder)
    remove_folder(temporary_folder)


def test_remove_path():
    temporary_path = mkstemp()[1]
    remove_path(temporary_path)
    remove_path(temporary_path)



from os.path import join

from invisibleroads_macros_disk import (
    check_absolute_path,
    check_path,
    check_relative_path,
    get_asset_path,
    get_file_hash,
    has_extension,
    is_matching_path,
    is_path_in_folder)
from invisibleroads_macros_disk.exceptions import PathValidationError
from pytest import raises

from conftest import A_FOLDER, B_FOLDER, FILE_NAME


def test_has_extension():
    assert has_extension('a.txt', ['.txt'])
    assert not has_extension('a.zip', ['.txt'])


def test_is_matching_path():
    assert is_matching_path('a.txt', ['*.txt'])
    assert not is_matching_path('a.zip', ['*.txt'])


def test_is_path_in_folder():
    assert is_path_in_folder(join(A_FOLDER, FILE_NAME), A_FOLDER)
    assert not is_path_in_folder(join(A_FOLDER, FILE_NAME), B_FOLDER)


def test_check_relative_path():
    path = join(A_FOLDER, FILE_NAME)
    assert check_relative_path(path, A_FOLDER) == FILE_NAME


def test_check_absolute_path():
    path = join(A_FOLDER, FILE_NAME)
    assert check_absolute_path(path, A_FOLDER) == path


def test_check_path():
    path = join(B_FOLDER, FILE_NAME)

    # Do not allow linked paths that resolve outside the folder
    with raises(PathValidationError):
        check_path(path, B_FOLDER)
    with raises(PathValidationError):
        check_path(path, B_FOLDER, trusted_folders=[B_FOLDER])

    # Allow linked paths if they resolve inside trusted folders
    check_path(path, B_FOLDER, trusted_folders=[A_FOLDER])


def test_get_file_hash():
    assert get_file_hash(join(A_FOLDER, FILE_NAME)).startswith('e')


def test_get_asset_path():
    package_name = 'invisibleroads_macros_disk'
    assert get_asset_path(package_name + ':__init__.py') == __import__(
        package_name).__file__
    assert get_asset_path('README.md') == 'README.md'



class InvisibleRoadsMacrosDiskError(Exception):
    pass


class PathValidationError(InvisibleRoadsMacrosDiskError):
    pass


class FileExtensionError(InvisibleRoadsMacrosDiskError):
    pass


class BadArchiveError(InvisibleRoadsMacrosDiskError):
    pass



from platformdirs import user_cache_dir


TEMPORARY_FOLDER = user_cache_dir(
    'invisibleroads-macros-disk', 'invisibleroads')
ARCHIVE_ZIP_EXTENSIONS = '.zip',
ARCHIVE_TAR_EXTENSIONS = '.tar.gz', '.tar.bz2', '.tar.xz'


CHUNK_SIZE_IN_BYTES = 2 ** 13
NAME_LENGTH = 2 ** 4



import re
import tarfile
from os import makedirs, remove
from os.path import exists, isdir, islink, join, splitext
from shutil import copyfile, rmtree
from urllib.request import urlretrieve as download
from zipfile import BadZipfile, ZipFile, ZIP_DEFLATED

from invisibleroads_macros_security import (
    get_hash,
    make_random_string)

from .constants import (
    ARCHIVE_TAR_EXTENSIONS,
    ARCHIVE_ZIP_EXTENSIONS,
    NAME_LENGTH,
    TEMPORARY_FOLDER)
from .exceptions import (
    BadArchiveError,
    FileExtensionError,
    InvisibleRoadsMacrosDiskError)
from .resolutions import (
    get_relative_path,
    has_extension,
    is_matching_path,
    walk_paths)


class TemporaryStorage(object):

    def __init__(
            self, base_folder=TEMPORARY_FOLDER, name_length=NAME_LENGTH,
            with_fixed_length=False):
        self.folder = make_random_folder(
            base_folder, name_length, with_fixed_length)

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        remove_folder(self.folder)


def archive_safely(source_folder, target_path=None, excluded_paths=None):
    'Archive without symbolic links. Specify archive extension in target_path.'
    if not target_path:
        target_path = source_folder + ARCHIVE_ZIP_EXTENSIONS[0]
    arguments = source_folder, target_path, excluded_paths
    if has_extension(target_path, ARCHIVE_ZIP_EXTENSIONS):
        archive_zip_safely(*arguments)
    elif has_extension(target_path, ARCHIVE_TAR_EXTENSIONS):
        archive_tar_safely(*arguments)
    else:
        archive_extensions = ARCHIVE_ZIP_EXTENSIONS + ARCHIVE_TAR_EXTENSIONS
        raise FileExtensionError({
            'target_path': 'must end in ' + ' or '.join(archive_extensions)})
    return target_path


def archive_zip_safely(source_folder, target_path=None, excluded_paths=None):
    'Archive folder as a zip archive without symbolic links'
    if not target_path:
        target_path = source_folder + ARCHIVE_ZIP_EXTENSIONS[0]
    if not has_extension(target_path, ARCHIVE_ZIP_EXTENSIONS):
        archive_extensions = ARCHIVE_ZIP_EXTENSIONS
        raise FileExtensionError({
            'target_path': 'must end in ' + ' or '.join(archive_extensions)})
    with ZipFile(
        target_path, 'w', ZIP_DEFLATED, allowZip64=True,
    ) as target_file:
        _compress_folder(source_folder, excluded_paths, target_file.write)
    return target_path


def archive_tar_safely(source_folder, target_path=None, excluded_paths=None):
    'Archive folder as a tar archive without symbolic links'
    if not target_path:
        target_path = source_folder + ARCHIVE_TAR_EXTENSIONS[0]
    if not has_extension(target_path, ARCHIVE_TAR_EXTENSIONS):
        archive_extensions = ARCHIVE_TAR_EXTENSIONS
        raise FileExtensionError({
            'target_path': 'must end in ' + ' or '.join(archive_extensions)})
    compression_format = splitext(target_path)[1].lstrip('.')
    with tarfile.open(
        target_path, 'w:' + compression_format, dereference=False,
    ) as target_file:
        _compress_folder(source_folder, excluded_paths, target_file.add)
    return target_path


def unarchive_safely(source_path, target_folder=None):
    'Unarchive folder without symbolic links'
    if has_extension(source_path, ARCHIVE_ZIP_EXTENSIONS):
        try:
            source_file = ZipFile(source_path, 'r')
        except BadZipfile:
            raise BadArchiveError({'source_path': 'is unreadable'})
        extension_expression = r'\.zip$'
        items = [_ for _ in source_file.infolist() if (
            _.external_attr >> 28) != 0xA]
    elif has_extension(source_path, ARCHIVE_TAR_EXTENSIONS):
        compression_format = splitext(source_path)[1].lstrip('.')
        try:
            source_file = tarfile.open(source_path, 'r:' + compression_format)
        except tarfile.ReadError:
            raise BadArchiveError({'source_path': 'is unreadable'})
        extension_expression = r'\.tar\.%s$' % compression_format
        items = [_ for _ in source_file.getmembers() if not _.issym()]
    else:
        archive_extensions = ARCHIVE_ZIP_EXTENSIONS + ARCHIVE_TAR_EXTENSIONS
        raise FileExtensionError({
            'source_path': 'must end in ' + ' or '.join(archive_extensions)})
    if not target_folder:
        target_folder = re.sub(extension_expression, '', source_path)
    for item in items:
        source_file.extract(item, target_folder)
    source_file.close()
    return target_folder


def make_enumerated_folder(base_folder, target_index=1):
    while True:
        target_folder = join(base_folder, str(target_index))
        try:
            makedirs(target_folder)
            break
        except FileExistsError:
            target_index += 1
    return target_folder


def make_random_folder(
        base_folder=TEMPORARY_FOLDER, name_length=NAME_LENGTH,
        with_fixed_length=False):
    while True:
        folder_name = make_random_string(name_length)
        folder = join(base_folder, folder_name)
        try:
            makedirs(folder)
            break
        except FileExistsError:
            if with_fixed_length:
                raise InvisibleRoadsMacrosDiskError(
                    f'could not make random folder in {base_folder}')
            name_length += 1
    return folder


def make_folder(folder):
    try:
        makedirs(folder)
    except FileExistsError:
        pass
    return folder


def remove_folder(folder):
    try:
        rmtree(folder)
    except FileNotFoundError:
        pass
    return folder


def remove_path(path):
    try:
        remove(path)
    except FileNotFoundError:
        pass
    return path


def cache_download(target_path, source_uri):
    if not exists(target_path):
        downloads_folder = make_folder(join(TEMPORARY_FOLDER, 'downloads'))
        source_path = join(downloads_folder, get_hash(source_uri))
        if exists(source_path):
            copyfile(source_path, target_path)
        else:
            download(source_uri, target_path)
            copyfile(target_path, source_path)
    return target_path


def _compress_folder(source_folder, excluded_paths, compress_path):
    for source_path in walk_paths(source_folder):
        if isdir(source_path) or islink(source_path):
            continue
        if is_matching_path(source_path, excluded_paths or []):
            continue
        target_path = get_relative_path(source_path, source_folder)
        compress_path(source_path, target_path)


make_unique_folder = make_random_folder



import fnmatch
from hashlib import blake2b
from os import walk
from os.path import abspath, expanduser, join, realpath, relpath
from pathlib import Path

from .constants import CHUNK_SIZE_IN_BYTES
from .exceptions import PathValidationError


def has_extension(path, extensions):
    for extension in extensions:
        if path.endswith(extension):
            return True
    return False


def is_matching_path(path, expressions):
    for expression in expressions:
        if fnmatch.fnmatch(path, expression):
            return True
    return False


def is_path_in_folder(path, folder):
    real_path = get_real_path(path)
    real_folder = get_real_path(folder)
    return real_path.startswith(real_folder)


def check_relative_path(path, folder, trusted_folders=None):
    check_path(path, folder, trusted_folders)
    return get_relative_path(path, folder)


def check_absolute_path(path, folder, trusted_folders=None):
    check_path(path, folder, trusted_folders)
    return get_absolute_path(path)


def check_path(path, folder, trusted_folders=None):
    real_path = get_real_path(path)
    for trusted_folder in trusted_folders or []:
        trusted_folder = get_real_path(trusted_folder)
        if real_path.startswith(trusted_folder):
            break
    else:
        real_folder = get_real_path(folder)
        if not real_path.startswith(real_folder):
            raise PathValidationError({
                'path': f'{real_path} is not in {real_folder}'})


def get_relative_path(path, folder):
    return relpath(expanduser(path), expanduser(folder))


def get_absolute_path(path):
    return abspath(expanduser(path))


def get_real_path(path):
    return realpath(expanduser(path))


def walk_paths(folder):
    for root_folder, folders, names in walk(folder):
        for name in folders + names:
            yield join(root_folder, name)


def get_file_hash(path, compute_hash=blake2b):
    with open(path, 'rb') as f:
        file_hash = compute_hash(usedforsecurity=False)
        while chunk := f.read(CHUNK_SIZE_IN_BYTES):
            file_hash.update(chunk)
    return file_hash.hexdigest()


def get_asset_path(asset_uri):
    asset_parts = asset_uri.split(':')
    if len(asset_parts) > 1:
        package_name, relative_path = asset_parts
        package_name = package_name.strip()
    else:
        package_name, relative_path = '', asset_parts[0]
    if package_name:
        package = __import__(package_name)
        package_folder = Path(package.__file__).parent
        path = str(package_folder / relative_path)
    else:
        path = relative_path
    return path
